/***************************************************
 * gldraw.c
 --------------------------------------------------*/
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <GL/glut.h>
#include "glgraph.h"
#include "bitmaps.c" // define mark[] and dot

#define MAX(a,b)     ((a > b) ? a : b)
#define MIN(a,b)     ((a < b) ? a : b)
    
/* common with glgraph.c */
extern void *setFont(char *fontname);
extern void *getFont(void);
extern int strwidth(char *str, char *fontname = NULL);
extern int strheight(char *str, char *fontname = NULL);

void putstring(double x, double y, int just, char *str)
{
    double
	xoffset = 0, yoffset = 0,
	swidth  = strwidth(str),
	sheight = strheight(str);
    switch (just) {
    case T_CENTER:     xoffset = -swidth/2; yoffset = -sheight/2; break;
    case T_LEFT:                            yoffset = -sheight/2; break;
    case T_UPPERLEFT:                       yoffset = -sheight;   break;
    case T_TOP:        xoffset = -swidth/2; yoffset = -sheight;   break;
    case T_UPPERRIGHT: xoffset = -swidth;   yoffset = -sheight;   break;
    case T_RIGHT:      xoffset = -swidth;   yoffset = -sheight/2; break;
    case T_LOWERRIGHT: xoffset = -swidth;                         break;
    case T_BOTTOM:     xoffset = -swidth/2;                       break;
    case T_LOWERLEFT:                                             break;
    }
    glRasterPos2f(x, y);
    glBitmap(1, 1, 0, 0, xoffset, yoffset, (unsigned char *)"\0");
    // glBitmap(0, 0, 0, 0, xoffset, yoffset, NULL); /* bug? in opengl32.dll */
    for (int i = 0; str[i]; i++) 
	glutBitmapCharacter(getFont(), str[i]);
}

void wprintf(double x, double y, int just, char *fmt, ...)
{
    va_list ap;
    char str[256];

    va_start(ap, fmt);
    vsprintf(str, fmt, ap);
    putstring(x, y, just, str);
    va_end(ap);
}

void line(double x0, double y0, double x1, double y1,
		 unsigned short mask = 0)
{
    if (mask) {
	glLineStipple(1, mask);
	glEnable(GL_LINE_STIPPLE);
	glBegin(GL_LINES); {
	    glVertex2d(x0, y0);
	    glVertex2d(x1, y1);
	}
	glEnd();
	glDisable(GL_LINE_STIPPLE);
    } else {
	glBegin(GL_LINES); {
	    glVertex2d(x0, y0);
	    glVertex2d(x1, y1);
	}
	glEnd();
    }
}

void linestrip(Point2d *pt, int N, unsigned short mask = 0)
{
    if (mask) {
	glLineStipple(1, mask);
	glEnable(GL_LINE_STIPPLE);
	glBegin(GL_LINE_STRIP);
	for (int i = 0; i < N; i++)
	    glVertex2d(pt[i].x, pt[i].y);
	glEnd();
	glDisable(GL_LINE_STIPPLE);
    } else {
	glBegin(GL_LINE_STRIP);
	for (int i = 0; i < N; i++)
	    glVertex2d(pt[i].x, pt[i].y);
	glEnd();
    }
}

void putDot(double x, double y, unsigned char *bitmap = NULL)
{
    glRasterPos2f(x, y);
    if (bitmap)
	glBitmap(8, 8, 3, 3, .0, .0, bitmap);
    else
	glBitmap(1, 1, 0, 0, .0, .0, (unsigned char *)"\255");
}

void putDots(Point2d *pt, int N, unsigned char *bitmap = NULL)
{
    if (bitmap) {
	for (int i = 0; i < N; i++) {
	    glRasterPos2f(pt[i].x, pt[i].y);
	    glBitmap(8, 8, 3, 3, .0, .0, bitmap);
	}
    } else {
	glBegin(GL_POINTS);
	for (int i = 0; i < N; i++)
	    glVertex2d(pt[i].x, pt[i].y);
	glEnd();
    }
}

void drawGrid(LocalWin *wi)
{
    char
	*XUnitText = pmString("XUnitText"),
	*YUnitText = pmString("YUnitText");
    int
	expX = (int)floor(log10(MAX(fabs(wi->UsrOrgX),fabs(wi->UsrOppX)))/3)*3,
	expY = (int)floor(log10(MAX(fabs(wi->UsrOrgY),fabs(wi->UsrOppY)))/3)*3;
    double
	Xincr = 50*wi->XUnitsPerPixel,  /* minimum spece of X-grid */
	Yincr = 25*wi->YUnitsPerPixel;  /* minimum spece of Y-grid */
    double
	dx = wi->XUnitsPerPixel,
	dy = wi->YUnitsPerPixel,
	x0 = wi->UsrOrgX, x1 = wi->UsrOppX,
	y0 = wi->UsrOrgY, y1 = wi->UsrOppY;

    /* X-Label */
    Point2d xuspot = {x1 + 10*dx, y0};          /* X-unit position */
    if (expX != 0) {
	xuspot.x += (strwidth(XUnitText) + 6*5)*dx;
	wprintf(xuspot.x, xuspot.y, T_RIGHT, "%s x 10", XUnitText);
	wprintf(xuspot.x, xuspot.y, T_LOWERLEFT, "%d", expX);
    } else { 
	wprintf(xuspot.x, xuspot.y, T_LEFT, "%s", XUnitText);
    }
    
    /* Y-Label */
    Point2d yuspot = {x0 - 40*dx, y1 + 20*dy};  /* Y-unit position */
    if (expY != 0) {
	yuspot.x += (strwidth(YUnitText) + 6*5)*dx;
	wprintf(yuspot.x, yuspot.y, T_RIGHT,"%s x 10", YUnitText);
	wprintf(yuspot.x, yuspot.y, T_LOWERLEFT,  "%d", expY);
    } else {
	wprintf(yuspot.x, yuspot.y, T_LEFT, "%s", YUnitText);
    }
    
    /* X-Grid */
    for (double x = initGrid(x0, Xincr, LogX()); x < x1; x = stepGrid()) {
	double y = y0 - 6*dy;
	if (LogX()) { /* log_X */
	    if (x == floor(x)) 
		wprintf(x, y, T_TOP, "%.0e", pow(10, x));
	    else
		wprintf(x, y, T_TOP, "%.2g", pow(10, x - floor(x)));
	} else {
	    wprintf(x, y, T_TOP, "%5.2f", x/pow(10, expX));
	}
	if (Ticks()) {
	    line(x, y0, x, y0 + 6*dy, 0xAAAA);
	    line(x, y1, x, y1 - 6*dy, 0xAAAA);
	} else {
	    line(x, y0, x, y1, 0xAAAA);
	}
    }
    
    /* Y-Grid */
    for (double y = initGrid(y0, Yincr, LogY()); y < y1; y = stepGrid()) {
	double x = x0 - 1*dx;
	if (LogY()) { /* log_Y */
	    if (y == floor(y))
		wprintf(x, y, T_RIGHT, "%.0e", pow(10, y));
	    else
		wprintf(x, y, T_RIGHT, "%.2g", pow(10, y - floor(y))); 
	} else {
	    wprintf(x, y, T_RIGHT, "%5.2f", y/pow(10, expY));
	}
	if (Ticks()) {
	    line(x0, y, x0 + 6*dx, y, 0xAAAA);
	    line(x1, y, x1 - 6*dx, y, 0xAAAA);
	} else {
	    line(x0, y, x1, y, 0xAAAA);
	}
    }
}

void gldraw(void)
{
    LocalWin *wi = getWinInfo();
    char *TitleText = pmString("TitleText");
    char 
	*LabelFont = pmString("LabelFont"),
	*TitleFont = pmString("TitleFont");
    double
	dx = wi->XUnitsPerPixel,
	dy = wi->YUnitsPerPixel,
	x0 = wi->UsrOrgX, x1 = wi->UsrOppX,
	y0 = wi->UsrOrgY, y1 = wi->UsrOppY,
	left   = x0 - wi->XOrgX*dx,
	right  = x1 + (wi->area_w - wi->XOppX)*dx,
        bottom = y0 - wi->XOrgY*dy,
	top    = y1 + (wi->area_h - wi->XOppY)*dy;
    GLfloat
	*bg_color = pmColor(ReverseVideo() ? "ForeGround" : "Background"),
	*fg_color = pmColor(ReverseVideo() ? "Background" : "Foreground"),
	*zg_color = pmColor(ReverseVideo() ? "Background" : "ZeroColor");
    GLfloat
	*color[MAXATTR] = {
	    pmColor("0.Color"), pmColor("1.Color"),
	    pmColor("2.Color"), pmColor("3.Color"),
	    pmColor("4.Color"), pmColor("5.Color"),
	    pmColor("6.Color"), pmColor("7.Color"),
	};

    /* set up window */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(left, right, bottom, top, -1.0, 1.0);
    glClearColor(bg_color[0], bg_color[1], bg_color[2], 0.0);
    glClear(GL_COLOR_BUFFER_BIT);

    /* Title */
    if (Grid()) {
	setFont(TitleFont);
	glColor3fv(fg_color);
	wprintf((x0 + x1)/2, y1 + 24*dy, T_BOTTOM, "%s", TitleText);
    }
	
    /* Grid */
    if (Grid()) {
	glColor3fv(fg_color);
	setFont(LabelFont);
	drawGrid(wi);
    }
    
    /* Axis */
    if (Axis()) {
	glColor3fv(zg_color);
	if (!LogY() && y0 < .0 && .0 < y1) line(x0, 0, x1, 0);
	if (!LogX() && x0 < .0 && .0 < x1) line(0, y0, 0, y1);
    }

    /* BBox */
    if (BoundBox()) {
	glColor3fv(fg_color);
	line(x0, y0, x1, y0, 0xAAAA);
	line(x1, y0, x1, y1, 0xAAAA);
	line(x1, y1, x0, y1, 0xAAAA);
	line(x0, y1, x0, y0, 0xAAAA);
    }

    /* legend */
    if (Grid()) {
	setFont(LabelFont);
	for (int idx = 0; idx < MAXSETS; idx++) {
	    if (PlotData[idx].list && *PlotData[idx].setName != '\0') {
		Point2d lspot = {x1 + 5*dx, y1 - 19*idx*dy}; /*legend top pos*/
		double lwidth = pmDouble("LegendWidth")*dx + 6*dx;
		glColor3fv(color[idx % MAXATTR]);
		if (Markers() || LargePixels())
		    putDot(lspot.x - dx, lspot.y,
			   Markers() ? mark[idx % MAXATTR] : dot);
		else if (NoLines())
		    line(lspot.x, lspot.y, lspot.x + lwidth, lspot.y, 0x1111);
		if (!NoLines()) 
		    line(lspot.x, lspot.y, lspot.x + lwidth, lspot.y);
		glColor3fv(fg_color);
		wprintf(lspot.x, lspot.y - 2*dy, T_UPPERLEFT,
			"%s", PlotData[idx].setName);
	    }
	}
    }

    /* curves */
    GLdouble clip[4][4] = {{ 1., 0., 0., -x0}, {-1., 0., 0.,  x1},
			   { 0., 1., 0., -y0}, { 0.,-1., 0.,  y1}};
    for (int i = 0; i < 4; i++) {
	glClipPlane((GLenum) (GL_CLIP_PLANE0 + i), clip[i]);
	glEnable((GLenum) (GL_CLIP_PLANE0 + i));
    }
    for (int idx = 0; idx < MAXSETS; idx++) {
	PointList *list = PlotData[idx].list;
	glColor3fv(color[idx % MAXATTR]);
	for ( ; list; list = list->next) {
	    if (!NoLines()) linestrip(list->pt, list->numPoints);
	    if (Markers() || LargePixels() || NoLines())
		putDots(list->pt, list->numPoints,
			Markers() ?  mark[idx % MAXATTR] :
			LargePixels() ? dot :
			/* else */     NULL);
	}
    }
    for (int i = 0; i < 4; i++) 
	glDisable((GLenum) (GL_CLIP_PLANE0 + i));
}
